#!/usr/bin/env python3
# coding=utf-8

from twitchez import command
from twitchez import conf
from twitchez.notify import notify

from shutil import which

import subprocess


def custom_cmd_check(type) -> tuple[bool, str]:
    if type == "stream":
        type_cmd = conf.setting("open_stream_cmd")
    elif type == "video":
        type_cmd = conf.setting("open_video_cmd")
    elif type == "extra":
        type_cmd = conf.setting("open_extra_cmd")
    else:
        type_cmd = conf.setting("open_stream_cmd")
    if type_cmd == "undefined":
        return False, ""
    executable = command.first_cmd_word(type_cmd)
    cmd_check = command.cmd_check(executable)
    return cmd_check, type_cmd


def custom_cmd(type_cmd, url) -> list:
    cmd = f"{type_cmd} {url}"
    return cmd.split()


def streamlink_cmd(url) -> list:
    streamlink_title_format = True
    quality = "best"  # hardcoded default
    cmd = "streamlink --quiet".split()
    if streamlink_title_format:
        # those are streamlink formatting variables
        title = '{author} - {category} -- {title}'
        cmd.append("--title")
        cmd.append(title)
    cmd_args: list = f"{url} {quality}".split()
    cmd.extend(cmd_args)
    return cmd


def mpv_cmd(url) -> list:
    cmd = f"mpv {url}"
    return cmd.split()


def raise_user_note():
    """raise exception for regular user without traceback."""
    a = "A program for opening url was not found at your 'PATH'."
    b = "You can install 'streamlink' and/or 'mpv' + 'youtube-dl/yt-dlp' and it will be working by default."
    c = "Also you can set your own program cmd via 'open_*_cmd = your cmd' in config."
    full_text = f"\n{a}\n{b}\n{c}\n"
    raise Exception(full_text)


def get_open_cmd(url, type):
    """Check & return cmd if executable is on PATH."""
    cmd = []
    # prefer custom open_cmd if set in config and found at PATH
    cmd_check, type_cmd = custom_cmd_check(type)
    if cmd_check and type_cmd:
        cmd = custom_cmd(type_cmd, url)
    # following are defaults and fallback if custom open cmd not set in config
    elif which("streamlink"):
        cmd = streamlink_cmd(url)
    elif which("mpv"):
        cmd = mpv_cmd(url)
    else:
        raise_user_note()
    return cmd


def open_url(url, type):
    """Open stream/video url with external/custom program."""
    cmd = get_open_cmd(url, type)
    if not cmd:
        return
    notify(url, "opening:", show_note=False)
    sub = subprocess.Popen
    sub(cmd,
        start_new_session=True,     # to not close video/stream after closing twitchez (POSIX only)
        stdin=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL)



#!/usr/bin/env python3
# coding=utf-8

from twitchez import STDSCR
from twitchez import keys
from twitchez import keys_help
from twitchez import render
from twitchez import tabs
from twitchez import thumbnails
from twitchez.keys import other_keys as k

from collections.abc import Callable
import curses


def set_curses_start_defaults():
    """Set curses start defaults."""
    curses.use_default_colors()
    curses.curs_set(0)   # Turn off cursor
    STDSCR.keypad(True)  # human friendly: curses.KEY_LEFT etc.


def wch() -> tuple[str, int, bool]:
    """Handle exceptions, and return character variables with explicit type."""
    try:
        wch = STDSCR.get_wch()
    except KeyboardInterrupt:  # Ctrl+c etc.
        thumbnails.draw_stop(safe=True)
        STDSCR.clear()
        curses.endwin()
        return "", 0, True  # fail/interrupt is True => break
    # explicit type conversion to be absolutely sure about character type!
    ci = int(wch) if isinstance(wch, int) else 0  # int else fallback to 0
    ch = str(wch)
    return ch, ci, False


def handle_resize(ci: int, redraw: Callable, redrawall: Callable):
    """Handle resize events, especially repeated resize -> simple redraw,
    when repeated resize stopped -> redraw everything including thumbnails."""
    if ci == curses.KEY_RESIZE:  # terminal resize event
        thumbnails.draw_stop()
        _rew = STDSCR.derwin(0, 0)  # resize event window (invisible)
        _rew.timeout(300)
        c = _rew.getch()
        # if next character is also a resize event
        if c == curses.KEY_RESIZE:
            _rew.timeout(150)
            # -> loop in simple redraw without thumbnails
            while c == curses.KEY_RESIZE:
                c = _rew.getch()
                STDSCR.clear()
                redraw()
        redrawall()
        return True
    return False


def show_pressed_chars(ch: str, ci: int):
    """Show last pressed key chars at the bottom-right corner."""
    h, w = STDSCR.getmaxyx()
    try:
        if ci != 0:
            STDSCR.insstr(h - 1, w - 8, f" ci:{ci} ")
        else:
            STDSCR.insstr(h - 1, w - 8, ch)
    except ValueError:  # bypass ValueError: embedded null character
        pass


def run(stdscr):
    global STDSCR
    STDSCR = stdscr  # override global STDSCR by the stdscr from wrapper
    page_dict = tabs.cpdict()  # last used page/tab
    page = render.Page(page_dict)

    set_curses_start_defaults()

    def redraw():
        """Reinitialize variables & redraw everything."""
        thumbnails.draw_stop()
        STDSCR.clear()
        h, w = STDSCR.getmaxyx()
        if h < 3 or w < 3:
            return
        page.draw()
        thumbnails.draw_start()

    redraw()  # draw once just before the loop start

    # Infinite loop to read every key press.
    while True:
        ch, ci, interrupt = wch()
        if interrupt:
            break
        if handle_resize(ci, page.draw, redraw):
            continue
        show_pressed_chars(ch, ci)

        if ch == k.get("quit"):
            break
        if ch == k.get("redraw"):
            page = render.Page(page_dict)
            redraw()
            continue
        if ch == k.get("redownload"):
            page = render.Page(page_dict, force_redownload=True)
            redraw()
            continue
        if ch == k.get("keys_help") or ci == curses.KEY_F1:
            keys_help.help()
            redraw()
            continue
        if ch == k.get("full_title"):
            STDSCR.clear()
            fbox = render.Boxes.drawn_boxes[0]
            # toggle full title drawing
            if not fbox.fulltitle:
                page.draw(fulltitle=True)
            else:
                page.draw()
            continue
        if ch in keys.bmark_keys.values():
            page_dict = keys.bmark_action(ch, page_dict)
            page = render.Page(page_dict)
            redraw()
            continue
        if keys.scroll(ch, page.draw, redraw):
            continue
        if ch in keys.tab_keys.values():
            page_dict = keys.tabs_action(ch, page_dict)
            page = render.Page(page_dict)
            redraw()
            continue
        if keys.yank(ch):
            continue
        if ch in keys.hint_keys.values():
            if keys.hints(ch):
                # redraw all including thumbnails
                redraw()
            else:
                # simple redraw without thumbnails
                STDSCR.clear()
                page.draw()
            continue
    # end of the infinite while loop


def main():
    try:
        curses.wrapper(run)
    finally:
        thumbnails.draw_stop(safe=True)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# coding=utf-8

from twitchez import HEADER_H
from twitchez import data
from twitchez import render
from twitchez import thumbnails
from twitchez import utils
from twitchez.utils import strws
from twitchez.tabs import tab_upd

from pathlib import Path


class Pages:

    def __init__(self, page_dict: dict, force_redownload=False):
        self.page_dict = page_dict
        self.page_name = page_dict["page_name"]
        self.cache_file_name = f"{strws(self.page_name)}.json"
        self.force_redownload: bool = force_redownload
        tab_upd(self.page_name, self.page_dict)  # => update tabs

    def cache_subdirs(self):
        """Return list of subdirs (to unpack them later as args)."""
        subdirs = []
        pd = self.page_dict
        ptype = pd.get("type", "streams")
        subdirs.append(ptype)
        if ptype == "videos":
            if "user_name" in pd:
                subdirs.append(strws(pd["user_name"]))
        if "category" in pd:
            subdirs.append(strws(pd["category"]))
        return subdirs

    def cache_path(self) -> Path:
        return data.cache_file_path(self.cache_file_name, *self.cache_subdirs())

    def update_cache(self) -> Path:
        return data.update_cache(self.cache_file_name, data.page_data(self.page_dict), *self.cache_subdirs())

    def read_cache(self) -> dict:
        return data.read_cache(self.cache_file_name, *self.cache_subdirs())

    def time_to_update_cache(self) -> bool:
        """Return True if path mtime > 5 mins from now.
        (default twitch API update time).
        """
        if self.force_redownload:
            self.force_redownload = False  # switch off to not redownload on each redraw call
            return True
        fnf = not Path(self.cache_path()).is_file()  # abbrev: file not found
        if fnf or utils.secs_since_mtime(self.cache_path()) > 300:
            return True
        else:
            return False

    def update_data(self) -> dict:
        """Update json data & return thumbnail paths."""
        subdirs = self.cache_subdirs()
        if self.time_to_update_cache():
            self.update_cache()
            json_data = self.read_cache()
            ids = data.get_entries(json_data, 'id')
            thumbnail_urls_raw = data.get_entries(json_data, 'thumbnail_url')
            thumbnail_paths = thumbnails.download_thumbnails(ids, thumbnail_urls_raw, *subdirs)
        else:
            # do not download thumbnails, find previously downloaded thumbnails paths
            json_data = self.read_cache()
            ids = data.get_entries(json_data, 'id')
            thumbnail_paths = thumbnails.find_thumbnails(ids, *subdirs)
        return thumbnail_paths

    def grid_func(self):
        """Return grid class object for prepared objects of thumbnails and boxes."""
        if thumbnails.text_mode():
            if self.time_to_update_cache():
                self.update_cache()
            thumbnail_paths = {}
        else:
            thumbnail_paths = self.update_data()
        json_data = self.read_cache()
        did = data.create_id_dict(json_data)  # dict with id as the key
        ids = list(did.keys())
        boxes = render.Boxes()
        grid = render.Grid(ids, self.page_name)
        for id, (x, y) in grid.coords.items():
            d = did[id]
            title = utils.tryencoding(d["title"])
            if "creator_name" in d:  # => clips
                # this is actually not login but name -> we do not need that anyway for clips
                user_login = d["broadcaster_name"]
                user_name = d["creator_name"]
            else:
                # used for composing stream url
                user_login = d["user_login"]
                user_name = d["user_name"]
            if not user_name:  # if user_name is empty (rare, but such case exist!)
                user_name = user_login
            # NOTE: videos DOES NOT HAVE game_name/category!
            if "game_name" in d:  # => live streams
                category = d["game_name"]
            elif "created_at" in d:  # => videos page
                category = utils.sdate(d["created_at"])
            elif "published_at" in d:
                category = utils.sdate(d["published_at"])
            else:
                category = ""
            if "viewer_count" in d:  # => live streams
                views = d["viewer_count"]
            elif "view_count" in d:  # => videos page
                views = d["view_count"]
            else:
                views = ""
            box = render.Box(user_login, user_name, title, category, x, y)
            if "url" in d:
                box.url = d["url"]  # videos have specific url
            if "duration" in d:
                box.duration = utils.duration(str(d["duration"]))
            box.viewers = str(views)
            if thumbnail_paths:
                box.img_path = thumbnail_paths[id]
                thumbnails.Thumbnail(id, thumbnail_paths[id], x, y + HEADER_H)
            boxes.add(box)
        return grid



#!/usr/bin/env python3
# coding=utf-8

from twitchez import STDSCR
from twitchez import command
from twitchez import conf
from twitchez import thumbnails

from os import environ
from shutil import which

import curses
import subprocess


select_cmd = conf.setting("select_cmd")
executable = command.first_cmd_word(select_cmd)
without_funcs = command.without_funcs(executable)
cmd_check = command.cmd_check(executable)


def dmenu_cmd() -> list:
    cmd = "dmenu -i -l 50"
    return cmd.split()


def fzf_cmd() -> list:
    cmd = environ.get("FZF_DEFAULT_COMMAND", "fzf")
    cmd += " --no-multi"
    return cmd.split()


def raise_user_note():
    """raise exception for regular user without traceback."""
    if without_funcs:
        return
    a = "A program for selecting of one line from all was not found at your 'PATH'."
    b = "You can install 'fzf' or 'dmenu' and it will be working by default."
    c = "Also you can set your own program cmd via 'select_cmd = your cmd' in config."
    d = "If you want to use this program without using it's interactive select functions,"
    e = "simply paste next line in your config:"
    f = "select_cmd = false"
    full_text = f"\n{a}\n{b}\n{c}\n{d}\n{e}\n{f}\n"
    raise Exception(full_text)


def get_select_cmd():
    """Check & return cmd if executable is on PATH."""
    cmd = []
    # prefer select_cmd if set in config and found at PATH
    if cmd_check:
        cmd = select_cmd.split()
    elif which("fzf"):
        cmd = fzf_cmd()
    elif which("dmenu"):
        cmd = dmenu_cmd()
    else:
        raise_user_note()
    return cmd


def iselect(multilinestr: str, fallback):
    """Interactive select of one line from all."""
    if without_funcs:
        return 130
    text = multilinestr.strip()
    cmd = get_select_cmd()
    # for fzf and similar console selectors working directly in terminal
    if cmd[0] != "dmenu" and cmd[0] != "rofi":
        STDSCR.refresh()  # fix: endwin() requires intervening screen update (new libncurses)
        curses.endwin()   # fix: hide application to be able to see selector after calling subprocess
        thumbnails.draw_stop()  # hide thumbnails, they will be redrawn in the next redraw() call.
    p = subprocess.run(cmd, input=text, text=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    sel = str(p.stdout).strip()
    if p.returncode == 1 or p.returncode == 130:
        # dmenu(1), fzf(130) => command was canceled (Esc)
        return 130
    elif p.returncode != 0:
        raise Exception(f"select cmd ERROR({p.returncode})\n{p.stderr}\n")
    # return fallback if input is not a substring of multilinestr
    if sel not in multilinestr:
        return fallback
    return sel



#!/usr/bin/env python3
# coding=utf-8

from twitchez import ENCODING
from twitchez import STDSCR
from twitchez import data
from twitchez import iselect
from twitchez import paged
from twitchez.notify import notify
import curses
import re


def inputwin(prompt: str) -> str:
    """Show input window at the last line of the stdscr window,
    return input str after pressing Enter key.
    """
    h, w = STDSCR.getmaxyx()
    win = curses.newwin(1, w // 3, h - 1, 0)
    win.addstr(0, 0, prompt, curses.A_REVERSE)
    win.refresh()
    curses.echo()
    # indent from the prompt by one character
    input = win.getstr(0, len(prompt) + 1)
    # convert bytes to string (remove the b prefix)
    decoded = str(input, ENCODING).strip()
    curses.noecho()
    win.erase()
    win.refresh()

    # Esc = b'\x1b', ^C = b'\x03'
    ignorebyte = ['\x1b', '\x03']
    ignorelist = ['/', '\\']
    ignorelist.extend(ignorebyte)

    # handle Esc or ^C from input as cancel command
    if any(_ in decoded for _ in ignorelist):
        notify("input was ignored!")
        return ""
    return decoded


def selected_category(fallback: dict) -> dict:
    input = inputwin("category:")
    if not input:
        return fallback
    mulstr = data.get_categories_terse_mulstr(input)
    selection = iselect.iselect(mulstr, 130)
    if selection == 130:
        return fallback
    id_pattern = re.compile(r"\[(\d+)\]$")
    sel_name = re.sub(id_pattern, "", selection).strip()
    sel_id = re.search(id_pattern, selection).group(1)
    category_id = sel_id
    category_name = sel_name
    page_dict = paged.stream(category_name, category_id)
    return page_dict


def selected_channel(video_type, fallback: dict) -> dict:
    input = inputwin("channel:")
    if not input:
        return fallback
    mulstr = data.get_channels_terse_mulstr(input)
    selection = iselect.iselect(mulstr, 130)
    if selection == 130:
        return fallback
    id_pattern = re.compile(r"\[(\d+)\]$")
    sel_id = re.search(id_pattern, selection).group(1)
    __sel_user = re.sub(id_pattern, "", selection).strip()
    sel_user = re.sub(r"^.*\s", "", __sel_user).strip()
    user_id = sel_id
    user_name = sel_user
    page_dict = paged.video(video_type, user_id, user_name)
    return page_dict


def select_page(fallback: dict) -> dict:
    """Interactive select of page to open, return page_dict of that page or fallback page."""
    msel = "category streams\nchannel videos\nfollowing live"
    main_sel = iselect.iselect(msel, 130)
    if main_sel == 130:  # handle cancel of the command
        return fallback
    if "following" in main_sel:
        return paged.following_live()
    elif "streams" in main_sel:
        return selected_category(fallback)
    # => videos page
    vtypes = "archive\nclips\nhighlight\nupload"
    video_type = iselect.iselect(vtypes, 130)
    if video_type == 130:
        return fallback
    return selected_channel(video_type, fallback)



#!/usr/bin/env python3
# coding=utf-8

from twitchez import STDSCR
from twitchez import bmark
from twitchez import data
from twitchez import search
from twitchez import tabs
from twitchez import thumbnails
from twitchez import utils
from twitchez.clip import clip
from twitchez.conf import key as ck
from twitchez.notify import notify
from twitchez.render import Boxes

from collections.abc import Callable
import curses

bmark_keys = {
    "bmark_add": ck("bmark_add"),
    "bmark_delete": ck("bmark_delete"),
    "bmark_open": ck("bmark_open"),
}

hint_keys = {
    "hint_clip_url": ck("hint_clip_url"),
    "hint_open_chat": ck("hint_open_chat"),
    "hint_open_stream": ck("hint_open_stream"),
    "hint_open_extra": ck("hint_open_extra"),
    "hint_open_video": ck("hint_open_video"),
}

scroll_keys = {
    "scroll_top": ck("scroll_top"),
    "scroll_bot": ck("scroll_bot"),
    "scroll_up": ck("scroll_up"),
    "scroll_down": ck("scroll_down"),
    "scroll_up_page": ck("scroll_up_page"),
    "scroll_down_page": ck("scroll_down_page")
}

tab_keys = {
    "tab_add": ck("tab_add"),
    "tab_delete": ck("tab_delete"),
    "tab_find": ck("tab_find"),
    "tab_next": ck("tab_next"),
    "tab_prev": ck("tab_prev"),
}

other_keys = {
    "quit": ck("quit"),
    "redraw": ck("redraw"),
    "redownload": ck("redownload"),
    "full_title": ck("full_title"),
    "keys_help": ck("keys_help"),
    "yank_urls": ck("yank_urls"),
    "yank_urls_page": ck("yank_urls_page"),
}


def rkt(timeout=500, fallback="") -> str:
    """Read key with timeout, without raising exception.
    (No exception if no input -> return fallback key or empty string.)
    """
    _rsw = STDSCR.derwin(0, 0)
    _rsw.timeout(timeout)
    try:
        c = str(_rsw.get_wch())
    except curses.error:  # "no input" etc.
        c = fallback
    return c


def bmark_action(ch: str, fallback: dict):
    """Bookmark action based on key."""
    page_dict = fallback
    if ch == bmark_keys.get("bmark_add"):
        bmark.bmark_add()
    elif ch == bmark_keys.get("bmark_delete"):
        bmark.bmark_del()
    elif ch == bmark_keys.get("bmark_open"):
        page_dict = bmark.bmark_open(fallback)
    return page_dict


def hints(ch: str):
    """Show box hints, and make some action based on key and hint.
    If terminal was resized while hints were being shown -> cancel & redraw all.
    """
    # get the initial sum to check later if terminal was resized
    xysum = utils.was_resized(0)

    boxes, hint = Boxes().show_boxes_hint()
    type = ""

    if ch == hint_keys.get("hint_clip_url"):
        type = "copy_url"
    elif ch == hint_keys.get("hint_open_chat"):
        type = "open_chat"
    elif ch == hint_keys.get("hint_open_stream"):
        type = "stream"
    elif ch == hint_keys.get("hint_open_video"):
        type = "video"
    elif ch == hint_keys.get("hint_open_extra"):
        type = "extra"

    if type == "copy_url":
        if not boxes.copy_url(hint) and not utils.was_resized(xysum):
            return False
    elif type == "open_chat":
        if not boxes.open_chat(hint) and not utils.was_resized(xysum):
            return False
    elif type:
        if not boxes.open_url(hint, type) and not utils.was_resized(xysum):
            return False

    # redraw all including thumbnails
    return True


def scroll_grid(redraw: Callable) -> str:
    """Scroll page grid based on the input key."""
    c = rkt(100)
    STDSCR.clear()
    grid = redraw()
    if c == scroll_keys.get("scroll_down"):
        grid.shift_index("down")
    elif c == scroll_keys.get("scroll_up"):
        grid.shift_index("up")
    elif c == scroll_keys.get("scroll_down_page"):
        grid.shift_index("down", page=True)
    elif c == scroll_keys.get("scroll_up_page"):
        grid.shift_index("up", page=True)
    elif c == scroll_keys.get("scroll_top"):
        grid.shift_index("top")
    elif c == scroll_keys.get("scroll_bot"):
        grid.shift_index("bot")
    return c


def scroll(ch: str, redraw: Callable, redrawall: Callable):
    """Scroll page, especially handle repeated scroll keys."""
    if ch in scroll_keys.values():
        thumbnails.draw_stop()
        c = rkt(100)
        curses.unget_wch(ch)
        # if no next input key after timeout -> scroll once
        if not c:
            c = scroll_grid(redraw)
        else:
            # if next character is also a scroll key
            # -> loop in simple redraw without thumbnails
            while c in scroll_keys.values():
                c = scroll_grid(redraw)
        redrawall()
        return True
    return False


def tabs_action(ch: str, fallback: dict):
    """Tabs actions."""
    if ch == tab_keys.get("tab_add"):
        page_dict = search.select_page(fallback)
    elif ch == tab_keys.get("tab_delete"):
        page_dict = tabs.delete_tab()
    elif ch == tab_keys.get("tab_find"):
        page_dict = tabs.find_tab()
    elif ch == tab_keys.get("tab_next"):
        page_dict, _ = tabs.next_tab()
    elif ch == tab_keys.get("tab_prev"):
        page_dict, _ = tabs.prev_tab()
    else:
        page_dict = fallback
    return page_dict


def yank_urls(full_page=False):
    """Yank urls of visible boxes or all urls of the page."""
    urls = ""
    if full_page:
        page_dict = tabs.cpdict()  # current page/tab
        json_data = data.page_data(page_dict)
        if "url" in json_data["data"][0]:
            page_urls = data.get_entries(json_data, "url")
            for url in page_urls:
                urls += f"{url}\n"
    else:
        for box in Boxes.drawn_boxes:
            urls += f"{box.url}\n"
    if urls:
        clip(urls)
    else:
        notify("This page does not have 'url' entries in json data.")


def yank(ch: str):
    if ch == other_keys.get("yank_urls") or ch == other_keys.get("yank_urls_page"):
        if ch == other_keys.get("yank_urls"):
            yank_urls()
        elif ch == other_keys.get("yank_urls_page"):
            yank_urls(full_page=True)
        return True
    return False



#!/usr/bin/env python3
# coding=utf-8

from twitchez import STDSCR
from twitchez import conf
import re


def get_hint_chars() -> str:
    return str(conf.setting("hint_chars"))


def total(items) -> tuple[int, int]:
    """Return (total_seq, hint_length) based on hint_chars,
    individual hint_length formula and len of items.
    total_seq = number of possible sequences,
    hint_length = [1-3] length of each hint.
    """
    hcl = len(get_hint_chars())
    sqr = hcl ** 2
    if hcl >= len(items):
        hint_length = 1
        total_seq = hcl
    elif sqr >= len(items):
        hint_length = 2
        total_seq = sqr
    else:
        hint_length = 3
        total_seq = sqr * 2 - hcl
    return total_seq, hint_length


def shorten_uniq_seq(out_seq):
    """Shorten all unique hint sequences and insert at the beginning."""
    tmp_seq = out_seq.copy()
    # shorten all sequences by one character
    for i, seq in enumerate(tmp_seq):
        wlc = seq[:-1]  # seq without last char
        tmp_seq.pop(i)
        tmp_seq.insert(i, wlc)
    # remove all elements that occurs more than once
    seq_set = set(tmp_seq)  # set() for less loop iterations
    for seq in seq_set:
        occurs = tmp_seq.count(seq)  # the number of times an element occurs
        if occurs > 1:
            while seq in tmp_seq:
                tmp_seq.remove(seq)
    if not tmp_seq:  # short unique sequences not found
        return out_seq
    # each letter associated with its position index
    order = {}
    for i, c in enumerate(get_hint_chars()):
        order[c] = i
    # compute the seq score by the order in which the letters appear in the sequence
    seq_score = {}
    for seq in tmp_seq:
        s1 = order[seq[0]]
        s2 = 0
        if len(seq) > 1:
            s2 = order[seq[1]]
        score = s1 + s2
        seq_score[seq] = score
    # dict of sequences sorted by the sequence score
    sorted_by_score = dict(sorted(seq_score.items(), key=lambda x: x[1]))
    sorted_short_seq = list(sorted_by_score.keys())
    sorted_short_seq.reverse()  # reverse() => we insert at the beginning
    # replace original seq by the shorter sequence as it occurs only once
    for sseq in sorted_short_seq:
        # original long seq found by the short seq
        llseq = [s for s in out_seq if re.search(f"^{sseq}.", s)]
        lseq = str(llseq[0])
        if lseq and lseq in out_seq:
            out_seq.remove(lseq)
        # insert all short sequences at the beginning
        out_seq.insert(0, sseq)
    return out_seq


def gen_hint_seq(items) -> list:
    """Generate from hint_chars list of unique sequences."""
    _, hint_length = total(items)
    hint_chars = get_hint_chars()

    # one letter length hints
    if hint_length == 1:
        return list(hint_chars)[:len(items)]

    # simple repeated values of hint_length
    repeated = []
    for c in hint_chars:
        # nn ee oo ... (if length_chars=2)
        repeated.append(c * hint_length)

    # make unique combinations of letters in strict order
    # generates sequence of 2 or 3 letter length hints
    combinations = []
    for r in repeated:
        new_seq = ""
        for ci in range(hint_length, 1, -1):
            pi = ci - 1
            for c in hint_chars:
                new_seq = r[:pi] + c + r[ci:]
                if new_seq in repeated:
                    continue  # skip
                if new_seq in combinations:
                    continue  # skip
                combinations.append(new_seq)

    hint_sequences = []
    hint_sequences.extend(repeated)
    hint_sequences.extend(combinations)
    # limit by the number of sequences that is enough for all items
    out_seq = hint_sequences[:len(items)]
    # NOTE: short seq are more convenient to type
    out_seq = shorten_uniq_seq(out_seq)
    # limit the number of sequences, strictly after shortening! (just in case)
    if len(out_seq) > len(items):
        return out_seq[:len(items)]
    else:
        return out_seq


def hint(items: list) -> list:
    """Return hint sequences for items."""
    return gen_hint_seq(items)


def find_seq(hints) -> str:
    """Input characters until only one hint sequence is found."""
    cinput = ""
    select = hints
    while len(select) > 1:
        c = str(STDSCR.get_wch())
        cinput += c
        select = [s for s in select if re.search(f"^{cinput}", s)]
    if not select:
        return ""
    if len(select) != 1:
        raise ValueError(f"len:({len(select)}) Only one item should be in the list:\n{select}")
    return str(select[0])



#!/usr/bin/env python3
# coding=utf-8

from twitchez import HEADER_H
from twitchez import STDSCR
from twitchez import conf
from twitchez import hints
from twitchez import open_chat
from twitchez import open_cmd
from twitchez import pages
from twitchez import utils
from twitchez.clip import clip
from twitchez.tabs import tab_names_ordered
from twitchez.thumbnails import container_size

from itertools import islice
from threading import Thread
from typing import TypeVar
import curses


SelfBoxes = TypeVar("SelfBoxes", bound="Boxes")


class Boxes:
    """Operate on list of Boxes"""
    boxlist = []
    drawn_boxes = []

    def add(self, obj):
        """Add box object to list."""
        self.boxlist.append(obj)

    def draw(self, parent, grid, fulltitle=False):
        """Draw boxes."""
        self.drawn_boxes.clear()
        stop = len(grid.coordinates())
        for box in islice(self.boxlist, stop):
            if fulltitle:
                box.fulltitle = True
            box.draw(parent)
            self.drawn_boxes.append(box)
        parent.refresh()
        self.boxlist.clear()

    def show_hints_boxes(self):
        """Show hints for visible/drawn boxes."""
        boxes = self.drawn_boxes
        hseq = hints.hint(boxes)
        for box, hint in zip(boxes, hseq):
            box.hint = hint
            box.show_hint()
        return hints.find_seq(hseq)

    def show_boxes_hint(self: SelfBoxes) -> tuple[SelfBoxes, str]:
        return self, self.show_hints_boxes()

    def get_box_attr_hint(self, hint, attr):
        """return attribute value of box object found by the hint."""
        boxes = self.drawn_boxes
        if not hasattr(boxes[0], attr):
            raise AttributeError(f"'{attr}' -> {boxes[0]} does not have such attribute!")
        for box in boxes:
            if getattr(box, "hint") == hint:
                return getattr(box, attr)
        raise Exception(f"value of ATTR:'{attr}' by the HINT:'{hint}' not found!")

    def copy_url(self, hint):
        if not hint:
            return
        value = self.get_box_attr_hint(hint, "url")
        clip(value)

    def open_url(self, hint, type):
        if not hint:
            return
        value = self.get_box_attr_hint(hint, "url")
        open_cmd.open_url(value, type)

    def open_chat(self, hint):
        if not hint:
            return
        value = self.get_box_attr_hint(hint, "user_login")
        open_chat.open_chat(value)


class Box:
    """Box with info about the stream/video inside the Grid."""
    w, h = container_size()
    last = h - 2  # last line of the box

    def __init__(self, user_login, user_name, title, category, x, y):
        self.user_login = user_login  # for composing url
        self.user_name = user_name
        self.title = utils.strclean(title)
        self.category = category
        self.x = x
        self.y = y
        self.url = f"https://www.twitch.tv/{self.user_login}"
        self.hint = ""
        self.img_path = ""
        self.viewers = ""
        self.duration = ""
        self.fulltitle = False

    def draw(self, parent):
        """Draw Box."""
        win = parent.derwin(self.h, self.w, self.y, self.x)
        win.addnstr(self.last, 0, f"{self.category}", self.w)
        if self.duration:
            duration = f"[{self.duration}]"
            rside = self.w - len(duration)
            win.addnstr(self.last - 1, 0, f"{self.user_name}", rside, curses.A_BOLD)
            win.addstr(self.last - 1, rside, duration)
        else:
            win.addnstr(self.last - 1, 0, f"{self.user_name}", self.w, curses.A_BOLD)
        if self.viewers:
            viewers = f" {self.viewers}"
            rside = self.w - len(viewers)
            win.addstr(self.last, rside, viewers, curses.A_BOLD)
        if self.fulltitle:
            max_len = int(self.w * 3)  # 3 box widths (lines)
            title = utils.word_wrap_title(self.title, self.w, max_len)
            try:
                win.addnstr(self.last - 2, 0, title, max_len)
            except Exception:
                win.box()
        else:
            title = utils.strtoolong(self.title, self.w)
            win.addnstr(self.last - 2, 0, title, self.w)

    def show_hint(self):
        """Create window with hint character."""
        if self.hint:  # if hint not empty -> show hint
            curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_RED)
            if len(self.hint) == 1:
                hint = f" {self.hint} "
            elif len(self.hint) == 2:
                hint = f" {self.hint}"
            else:
                hint = f"{self.hint}"
            lh = len(hint)
            win = curses.newwin(1, lh + 1, self.y + self.h - 1, self.x + self.w - lh)
            win.addstr(hint, curses.color_pair(1))
            win.refresh()


class Grid:
    """Grid of boxes inside the Window."""
    w, h = container_size()

    def __init__(self, key_list: list, page_name: str):
        self.key_list = key_list
        self.page_name = page_name
        self.__ba = BodyArea()
        self.area_cols = self.__ba.cols
        self.area_rows = self.__ba.rows
        self.key_start_index = self.index()
        self.coords = self.coordinates()

    def capacity(self) -> tuple[int, int, int]:
        """Return - how many boxes can fit in: (cols, rows, total)."""
        cols = self.area_cols // self.w
        rows = self.area_rows // self.h
        total = cols * rows
        return cols, rows, total

    def spacing(self, cols, rows) -> tuple[int, int]:
        """Calculate even spacing between grid elements.
        returns spacing: cols, rows.
        """
        if cols < 1:
            c = 0
        else:
            c = int(self.area_cols - self.w * cols) // cols
        if rows < 1:
            r = 0
        else:
            r = int(self.area_rows - self.h * rows) // rows
        return c, r

    def index(self, start_index="") -> int:
        """Set/Get initial grid index."""
        if str(start_index):  # str -> to check if not empty (even 0 value)
            index = int(start_index)
            conf.tmp_set("grid_index", index, self.page_name)
        else:
            index = int(conf.tmp_get("grid_index", 0, self.page_name))
            if not index or index < 0:
                index = 0
                conf.tmp_set("grid_index", index, self.page_name)
        return index

    def shift_index(self, dir="down", page=False) -> int:
        """Shift value of the key start index."""
        cols, _, total = self.capacity()
        elems_total = len(self.key_list)
        remainder = elems_total % cols
        if remainder == 0:
            end_of_page = elems_total - total
        else:
            end_of_page = elems_total - total - remainder + cols
        if elems_total <= total:
            start_index = 0
        else:
            grid_index = self.index()
            if dir == "top":
                start_index = 0
            elif dir == "bot":
                start_index = end_of_page
            elif page:
                if dir == "down":
                    start_index = grid_index + total
                else:
                    start_index = grid_index - total
            else:
                if dir == "down":
                    start_index = grid_index + cols
                else:
                    start_index = grid_index - cols
            if start_index < 0:
                start_index = 0
            elif start_index > end_of_page:
                start_index = end_of_page
        self.index(str(start_index))
        return start_index

    def coordinates(self) -> dict:
        """Return dict with: tuple(X, Y) values where each key_list element is the key."""
        initial_x, initial_y = 0, 0
        cols, rows, total = self.capacity()
        total += self.key_start_index  # for scrolling
        scols, srows = self.spacing(cols, rows)
        # for more even spacing from both sides
        sc = scols // 2
        sr = srows // 2
        x = initial_x + sc
        y = initial_y + sr
        current_col = 1
        coordinates = {}
        for key in islice(self.key_list, self.key_start_index, total):
            if cols > 2:
                x += sc
            coordinates[key] = (x, y)
            if current_col < cols:
                current_col += 1
                x += sc + self.w
            else:
                current_col = 1
                x = initial_x + sc
                y += sr + self.h
        return coordinates


class Page:
    """Page which renders everything."""

    def __init__(self, page_dict, force_redownload=False):
        self.pages_class = pages.Pages(page_dict, force_redownload)
        self.page_name = self.pages_class.page_name
        self.grid_func = self.pages_class.grid_func
        self.loaded = False

    def loading(self):
        """Simple animation to show that something is being done (Page loading).
        Currently the animation cycle is very short and ends even
        if the loading is not yet finished, '*' - static indicator of this.

        This is done intentionally to not ruin everything
        if raise() or crash occurred while thread is not yet finished and etc.
        The animation is short to quickly return to the terminal if an error is raised.
        """
        # NOTE: currently ANIMATION introduces extra wait time during draw() calls
        # on simple and fast operations like redraw() => so animation is disabled.
        # It does not feel like the fancy animation is worth it.
        ANIMATION = False

        def animation():
            """Animation length is intentionally short."""
            chars = "-\\|/"  # animation chars
            for _ in range(8):
                for c in chars:
                    win.insstr(c)
                    win.refresh()
                    curses.napms(25)
                    # finish animation right now!
                    if self.loaded:
                        return

        def anima_thread():
            t = Thread(target=animation())
            t.start()
            t.join()

        try:
            win = curses.newwin(1, 1, 0, 0)
        except Exception:
            return

        try:
            if ANIMATION:
                anima_thread()
            else:
                win.insstr("*")
        finally:
            if self.loaded:
                win.erase()
            else:
                # leave a static indicator about not yet finished loading
                # it probably will be cleared by some clear of the screen
                # so we do not bother much about clearing of the static indicator :)
                win.insstr("*")
            win.refresh()

    def draw_header(self):
        """Draw page header."""
        indent = 2  # indent from side
        indent_between = " "
        separator = "|"  # separator between tabs
        between_tabs = indent_between + separator + indent_between
        logo = "[twitchez]"
        c_page = self.page_name  # current page name
        _, w = STDSCR.getmaxyx()
        head = STDSCR.derwin(HEADER_H - 1, w, 0, 0)
        other_tabs = ""
        # tab order where current page is always first in list (to look as carousel)
        taborder = []
        tnames = tab_names_ordered()
        cpni = tnames.index(c_page)
        taborder.extend(tnames[cpni:])
        taborder.extend(tnames[:cpni])
        for tab in taborder:
            if tab == c_page:
                continue  # skip current tab
            # indent_between with separator for each additional tab page
            other_tabs += between_tabs + tab
        icp = indent + len(c_page)  # width of current page with indent
        all_tabs = indent + len(c_page + other_tabs)
        if w > all_tabs:  # if we can fit all tabs
            head.addnstr(0, indent, c_page, len(c_page), curses.A_REVERSE)  # current Tab page
            head.addnstr(0, icp, other_tabs, len(other_tabs))
            if w > all_tabs + len(logo) + 1:  # if window have enough width for logo
                wllimit = w - len(logo) - indent
                head.addnstr(0, wllimit, logo, len(logo), curses.A_BOLD)
        else:
            # crop tabs visually that do not fit into the window
            # > character signalize about cropping (existence of other tabs)
            if w > indent:
                wclimit = w - indent - 1
                c_page = c_page[:wclimit - 1] + ">"
                head.addnstr(0, indent, c_page, wclimit, curses.A_REVERSE)
            if w > icp:
                wolimit = w - icp - 1
                other_tabs = other_tabs[:wolimit - 1] + ">"
                head.addnstr(0, icp, other_tabs, wolimit)
        head.refresh()
        return head

    def draw_body(self, grid, fulltitle=False):
        """Draw page body."""
        body = BodyArea().window()
        if fulltitle:
            Boxes().draw(body, grid, fulltitle)
        else:
            Boxes().draw(body, grid)
        return body

    def draw(self, fulltitle=False):
        """return grid and draw full page."""
        self.loading()
        grid = self.grid_func()
        self.draw_body(grid, fulltitle)
        self.draw_header()
        self.loaded = True  # finish animation of loading if not yet ended
        return grid


class BodyArea:
    """Body area size of the window in the terminal cells."""

    def __init__(self):
        self.rows, self.cols = STDSCR.getmaxyx()
        self.rows = self.rows - HEADER_H

    def window(self):
        """Create & return body window."""
        return STDSCR.derwin(self.rows, self.cols, HEADER_H, 0)



#!/usr/bin/env python3
# coding=utf-8

from twitchez import conf

from shutil import which


def first_cmd_word(cmd: str) -> str:
    return cmd.split()[0]


def without_funcs(executable: str) -> bool:
    """return True if funcs are disabled via cmd in settings."""
    if "false" in executable.lower():
        return True
    return False


def executable_check(executable: str) -> bool:
    """return True if executable(first word from cmd) set in config and found at PATH."""
    if executable != "undefined":
        if which(executable):
            return True
    return False


def cmd_check(executable: str) -> bool:
    """return True if funcs are enabled & checks of executable are passed."""
    if not without_funcs(executable):
        if executable_check(executable):
            return True
    return False


def conf_cmd_check(conf_cmd: str) -> tuple[bool, str]:
    """check, return cmd from config."""
    cmd = conf.setting(conf_cmd)
    executable = first_cmd_word(cmd)
    cmd_ok = cmd_check(executable)
    return cmd_ok, cmd




#!/usr/bin/env python3
# coding=utf-8

from pathlib import Path
from requests import get
from twitchez import fs
import json


def validate_data(d :dict):
    """Check and handle status code if found in data."""
    status = 0 # initial status code (ALL OK)
    if "status" in d:
        status = int(d["status"])
    if not status:
        return
    # ^ if status code not 0 -> status code processing
    if status == 401: # Invalid OAuth token
        fs.private_data_path(recreate=True)
        # message to the user
        a = "Invalid OAuth token! (probably the old token has expired)"
        b = "Launch application again to generate new OAuth token."
        raise Exception(f"{str(d)}\n^{a}\n{b}")
    else:
        a = "returned request data:"
        b = "*** Unhandled status code! ***"
        raise Exception(f"{a}\n{str(d)}\n{b} ({status})")


def write_private_data(user_id, access_token, client_id):
    """Write private data to file for using in further requests."""
    file_path = fs.private_data_path(recreate=True)
    data = {
        "u_id": user_id,
        "token": access_token,
        "c_id": client_id
    }
    with open(file_path, "w") as file:
        json.dump(data, file, indent=4)


def get_private_data(key) -> str:
    """Get value by the key from .private file."""
    file_path = fs.private_data_path()
    with open(file_path, "r") as file:
        data = json.load(file)
    return data[key]


def cache_file_path(file_name, *subdirs) -> Path:
    """Get cache file path by file name, optionally supports subdirs."""
    if subdirs:
        path = Path(fs.get_cache_dir(), *subdirs)
        path.mkdir(parents=True, exist_ok=True)
    else:
        path = fs.get_cache_dir()
    return Path(path, file_name)


def update_cache(file_name, json_data, *subdirs) -> Path:
    """Update json file from cache and return file path."""
    if subdirs:
        file_path = cache_file_path(file_name, *subdirs)
    else:
        file_path = cache_file_path(file_name)
    data = json.dumps(json_data, indent=2)
    with open(file_path, "w") as file:
        file.write(data)
    return file_path


def read_cache(file_name, *subdirs) -> dict:
    """Read json file from cache and return data."""
    if subdirs:
        file_path = cache_file_path(file_name, *subdirs)
    else:
        file_path = cache_file_path(file_name)
    with open(file_path, "r") as file:
        data = json.load(file)
    return data


def get_entries(json_data, key, root_key='data') -> list:
    """Create and return list of values from json data where all entries found by key."""
    found = []
    for entry in json_data[root_key]:
        found.append(entry[key])
    return found


def create_id_dict(json_data) -> dict:
    """Create and return dict with id as the key."""
    streams = {}
    ids = get_entries(json_data, 'id')
    for stream, id in zip(json_data['data'], ids):
        streams[id] = stream
    return streams


def following_live_data() -> dict:
    """Return data of user 'following live channels' page."""
    u_id = get_private_data("u_id")    # user_id
    token = get_private_data("token")  # auth token
    c_id = get_private_data("c_id")    # client-Id of this program
    url = f"https://api.twitch.tv/helix/streams/followed?user_id={u_id}"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": c_id
    }
    r = get(url, headers=headers)
    d = r.json()
    validate_data(d)
    return d


def get_categories(query: str) -> list:
    """Returns a list of categories that match the query via name either entirely or partially."""
    first = 100  # Maximum number of objects to return. (Twitch API Maximum: 100)
    token = get_private_data("token")
    c_id = get_private_data("c_id")
    url = f"https://api.twitch.tv/helix/search/categories?first={first}&query={query}"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": c_id
    }
    r = get(url, headers=headers)
    d = r.json()
    validate_data(d)
    return d["data"]


def get_categories_terse_data(query: str) -> dict:
    categories = get_categories(query)
    terse_info = {}
    for c in categories:
        # dict key is id = tuple of ...
        terse_info[c["id"]] = c["name"], c["box_art_url"]
    return terse_info


def get_categories_terse_mulstr(query: str) -> str:
    """Return multiline string with terse categories data. (for interactive select)"""
    d = get_categories_terse_data(query)
    mstr = ""
    names = []
    for v in d.values():
        name, _ = v
        names.append(name)
    maxlen = len(max(names, key=len))  # max length of longest string in list
    for id, v in d.items():
        name, _ = v
        mstr += f"{str(name):<{int(maxlen)}} [{id}]\n"
    return mstr.strip()  # to remove blank line


def category_data(category_id) -> dict:
    """Return json data for streams in certain category."""
    first = 100  # Maximum number of objects to return. (Twitch API Maximum: 100)
    token = get_private_data("token")
    c_id = get_private_data("c_id")
    url = f"https://api.twitch.tv/helix/streams?first={first}&game_id={category_id}"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": c_id
    }
    r = get(url, headers=headers)
    d = r.json()
    validate_data(d)
    return d


def get_channels(query: str, live_only=False) -> dict:
    """Returns a list of channels that match the query via channel name.
    (users who have streamed within the past 6 months)
    """
    first = 5  # Maximum number of objects to return. (Twitch API Maximum: 100)
    token = get_private_data("token")
    c_id = get_private_data("c_id")
    url = f"https://api.twitch.tv/helix/search/channels?first={first}&live_only={str(live_only)}&query={query}"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": c_id
    }
    r = get(url, headers=headers)
    d = r.json()
    validate_data(d)
    return d["data"]


def get_channels_terse_data(query: str, live_only=False) -> dict:
    """Return id dict with channels: (broadcaster_login, display_name, profile_image_url) only."""
    channels = get_channels(query, live_only)
    terse_info = {}
    for ch in channels:
        # dict key is channel id = tuple of ...
        terse_info[ch["id"]] = ch["broadcaster_login"], ch["display_name"], ch["thumbnail_url"]
    return terse_info


def get_channels_terse_mulstr(query: str, live_only=False) -> str:
    """Return multiline string with terse channels data. (for interactive select)"""
    d = get_channels_terse_data(query, live_only)
    mstr = ""
    maxlen = 15
    for id, ch in d.items():
        login, name, _ = ch
        mstr += f"{str(login):<{maxlen}} {str(name):<{maxlen}} [{id}]\n"
    return mstr.strip()  # to remove blank line


def get_channel_videos(user_id, type="all") -> dict:
    """Gets videos information by user ID."""
    first = 100  # Maximum number of objects to return. (Twitch API Maximum: 100)
    token = get_private_data("token")
    c_id = get_private_data("c_id")
    url = f"https://api.twitch.tv/helix/videos?type={type}&first={first}&user_id={user_id}"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": c_id
    }
    r = get(url, headers=headers)
    d = r.json()
    validate_data(d)
    return d


def get_channel_clips(broadcaster_id) -> dict:
    """Gets clips information by broadcaster ID."""
    first = 100  # Maximum number of objects to return. (Twitch API Maximum: 100)
    token = get_private_data("token")
    c_id = get_private_data("c_id")
    url = f"https://api.twitch.tv/helix/clips?first={first}&broadcaster_id={broadcaster_id}"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": c_id
    }
    r = get(url, headers=headers)
    d = r.json()
    validate_data(d)
    return d


def page_data(page_dict) -> dict:
    """Get and return page data based on page_dict."""
    pd = page_dict
    ptype = pd.get("type", "streams")
    if ptype == "videos":
        if pd["category"] == "clips":
            json_data = get_channel_clips(pd["user_id"])
        else:
            json_data = get_channel_videos(pd["user_id"], pd["category"])
    else:
        if pd["category"] == "Following Live":
            json_data = following_live_data()
        else:
            json_data = category_data(pd["category_id"])
    return json_data



#!/usr/bin/env python3
# coding=utf-8

from os import environ
from pathlib import Path
from tempfile import gettempdir


def set_owner_only_permissions(path: Path) -> Path:
    """Set owner only path permissions."""
    if path.is_dir():
        path.chmod(0o700, follow_symlinks=True)
    else:
        path.chmod(0o600, follow_symlinks=True)
    return path


def private_data_path(recreate=False) -> Path:
    """Check that the .private file exists, if not -> create empty file.
    Also set r+w dir & file permissions to owner only & return path to file.
    """
    private_dir = get_data_dir(".private")       # create dir if not exist
    file_path = Path(private_dir, ".private")
    # remove file with old private data for authentication
    if recreate and file_path.exists():
        file_path.unlink(missing_ok=True)
    if not file_path.exists():
        file_path.touch(exist_ok=True)           # create empty file
        set_owner_only_permissions(private_dir)  # set dir permissions
        set_owner_only_permissions(file_path)    # set file permissions
    return file_path


def get_cache_dir() -> Path:
    """Check ENV variables, create cache dir and return it's path."""
    dirname = "twitchez"
    if "TWITCHEZ_CACHE_DIR" in environ:
        cache_home = environ["TWITCHEZ_CACHE_DIR"]
    elif "XDG_CACHE_HOME" in environ:
        cache_home = environ["XDG_CACHE_HOME"]
    else:
        cache_home = Path(Path.home(), ".cache")
    cache_dir = Path(cache_home, dirname)
    # create cache_dir if not exist
    Path(cache_dir).mkdir(parents=True, exist_ok=True)
    return cache_dir


def get_data_dir(*subdirs) -> Path:
    """Return path to data dir and create optional subdirs if they doesn't already exist."""
    dirname = "twitchez"
    if "TWITCHEZ_DATA_DIR" in environ:
        data_home = environ["TWITCHEZ_DATA_DIR"]
    elif "XDG_DATA_HOME" in environ:
        data_home = environ["XDG_DATA_HOME"]
    else:
        data_home = Path(Path.home(), ".local", "share")
    if not subdirs:
        data_path = Path(data_home, dirname)
    else:
        data_path = Path(data_home, dirname, *subdirs)
    # create data_path dirs if not exist
    Path(data_path).mkdir(parents=True, exist_ok=True)
    return data_path


def get_tmp_dir(*subdirs) -> Path:
    """Return path to tmp dir and create optional subdirs if they doesn't already exist."""
    dirname = "twitchez"
    if not subdirs:
        tmp_dir_path = Path(gettempdir(), dirname)
    else:
        tmp_dir_path = Path(gettempdir(), dirname, *subdirs)
    Path(tmp_dir_path).mkdir(parents=True, exist_ok=True)
    return tmp_dir_path


def get_user_conf_dir() -> Path:
    """Check ENV variables, get user config dir and return it's path."""
    dirname = "twitchez"
    if "XDG_CONFIG_HOME" in environ:
        config_home = environ["XDG_CONFIG_HOME"]
    else:
        config_home = Path(Path.home(), ".config")
    config_dir = Path(config_home, dirname)
    return config_dir



#!/usr/bin/env python3
# coding=utf-8

FLPN = "Following Live"


def following_live() -> dict:
    """Following Live page dict."""
    return {
        "type": "streams",
        "category": FLPN,
        "page_name": FLPN
    }


def stream(category_name: str, category_id: str) -> dict:
    """Stream page dict."""
    return {
        "type": "streams",
        "category": category_name,
        "page_name": category_name,
        "category_id": category_id
    }


def video(video_type: str, user_id: str, user_name: str) -> dict:
    """Video page dict."""
    return {
        "type": "videos",
        "category": video_type,
        "page_name": f"{user_name} ({video_type})",
        "user_name": user_name,
        "user_id": user_id
    }



#!/usr/bin/env python3
# coding=utf-8

from ast import literal_eval
from pathlib import Path
from twitchez import conf
from twitchez import fs
from twitchez.iselect import iselect
from twitchez.tabs import find_tab, cpname_set, tab_add_new

bmsf = Path(fs.get_data_dir("data"), "bookmarks").resolve().as_posix()
SECT = "BMARKS"


def bmark_list() -> list:
    """Return list of bookmark (name, dict) tuples."""
    return conf.dta_list(SECT, bmsf)


def bmark_names() -> list:
    """Return list of bookmark names."""
    return [bname for bname, _ in bmark_list()]


def bmark_save(page_name: str, page_dict: dict):
    """Save bookmark."""
    conf.dta_set(page_name, page_dict, SECT, bmsf)


def bmark_add():
    """Find tab and save as bookmark."""
    page_dict = find_tab({})
    # handle cancel of the command
    if not page_dict:
        return
    page_name = page_dict.get("page_name")
    bmark_save(page_name, page_dict)


def bmark_find() -> tuple[str, dict]:
    """Find and return (name, dict) tuple of the selected bookmark."""
    bnames = bmark_names()
    mulstr = "\n".join(bnames)
    bname = iselect(mulstr, 130)
    # handle cancel of the command
    if (bname == 130):
        return "", {}
    # get dict by the key
    bdict = {}
    for key, val in bmark_list():
        if (key == bname):
            bdict = dict(literal_eval(val))
            break
    return bname, bdict


def bmark_del():
    """Delete bookmark by the name."""
    bname, _ = bmark_find()
    if (not bname):
        return
    conf.dta_rmo(bname, SECT, bmsf)


def bmark_open(fallback: dict) -> dict:
    """Open bookmark by the name."""
    bname, bdict = bmark_find()
    if (not bname or not bdict):
        return fallback
    tab_add_new(bname)
    cpname_set(bname)
    return bdict



#!/usr/bin/env python3
# coding=utf-8

from pathlib import Path
import curses

__version__ = "0.1.0"
__license__ = "GPLv3"
__description__ = "twitchez - TUI client for twitch.tv with thumbnails support that works right in your terminal"
__url_repository__ = "https://github.com/WANDEX/twitchez"
__url_bug_reports__ = "https://github.com/WANDEX/twitchez/issues"
__url_project__ = __url_repository__
__author__ = "WANDEX"

# Constants
ENCODING = "utf-8"
HEADER_H = 2
TWITCHEZDIR = Path(__file__).parent.resolve()

# NOTE: STDSCR defined here for ease of reuse, to be able to see actual curses funcs
# and clarify variable type for: interpreter, diagnostic messages, developer, etc.
# we are initializing the actual application later with the curses.wrapper()
# then we override this global constant by the actual stdscr right after initialization
try:
    STDSCR = curses.initscr()
    # all of the following to safely terminate temporary initialized curses application
    # and restore the terminal to default settings and operating mode.
    STDSCR.keypad(False)
    curses.echo()
    curses.nocbreak()
    STDSCR.refresh()  # fix: endwin() requires intervening screen update (new libncurses)
    curses.endwin()
except Exception:
    try:
        curses.endwin()
    except Exception:
        pass



#!/usr/bin/env python3
# coding=utf-8

from shutil import which
from twitchez import ENCODING
from twitchez import command
from twitchez import conf
from twitchez.notify import notify
import subprocess

clip_cmd = conf.setting("clip_cmd")
executable = command.first_cmd_word(clip_cmd)
without_funcs = command.without_funcs(executable)
cmd_check = command.cmd_check(executable)


def xclip_cmd() -> list:
    cmd = "xclip -in -selection clipboard"
    return cmd.split()


def xsel_cmd() -> list:
    cmd = "xsel -i --clipboard"
    return cmd.split()


def raise_user_note():
    """raise exception for regular user without traceback."""
    if without_funcs:
        return
    a = "A program for copying content to clipboard was not found at your 'PATH'."
    b = "You can install 'xclip' and it will be working by default."
    c = "Also you can set your own program cmd via 'clip_cmd = your cmd' in config."
    d = "If you want to use this program without using it's clipboard functions,"
    e = "simply paste next line in your config:"
    f = "clip_cmd = false"
    full_text = f"\n{a}\n{b}\n{c}\n{d}\n{e}\n{f}\n"
    raise Exception(full_text)


def get_clip_cmd(show_note: bool) -> list:
    """Check & return cmd if executable is on PATH."""
    cmd = []
    # prefer clip_cmd if set in config and executable found at PATH
    if cmd_check:
        cmd = clip_cmd.split()
    elif which("xclip"):
        cmd = xclip_cmd()
    elif which("xsel"):
        cmd = xsel_cmd()
    else:
        if show_note:
            raise_user_note()
        else:
            return []
    return cmd


def clip(content: str, show_note=True):
    """Copy content to clipboard."""
    if without_funcs:
        return
    cmd = get_clip_cmd(show_note=show_note)
    if not cmd:
        return
    text = content.strip()
    p = subprocess.Popen(cmd, stdin=subprocess.PIPE, close_fds=True)
    p.communicate(input=text.encode(ENCODING))
    p.wait()  # wait for process to finish
    if p.returncode == 0:
        notify(text, "C:", show_note=show_note)
    else:
        notify(f"ERROR({p.returncode}): probably malformed cmd!",
               "NOT copied:", error=True, show_note=show_note)



#!/usr/bin/env python3
# coding=utf-8

from random import randint
from twitchez.clip import clip
from twitchez.data import write_private_data
import requests


def generate_nonce(length=8):
    """Generate pseudorandom number."""
    return ''.join([str(randint(0, 9)) for _ in range(length)])


def get_user_id(token, c_id):
    """Get user id by access token."""
    url = "https://api.twitch.tv/helix/users"
    headers = {
        "Authorization": f"Bearer {token}",
        "Client-Id": c_id
    }
    try:
        r = requests.get(url, headers=headers)
    except Exception as err:
        raise Exception(err)
    return(r.json()['data'][0]['id'])


def get_auth_token():
    """Read more here:
    'https://dev.twitch.tv/docs/authentication/getting-tokens-oauth/#oauth-implicit-code-flow'
    """
    client_id = "dadsrpg93f0tvvq8zhbno69m2e3spr"  # this application client id
    redirect_uri = "https://localhost"
    scope = "user:read:follows"
    state = generate_nonce()  # against simple CSRF attacks
    url = "".join((
        "https://id.twitch.tv/oauth2/authorize"
        f"?client_id={client_id}"
        f"&redirect_uri={redirect_uri}"
        "&response_type=token"
        f"&scope={scope}"
        f"&state={state}"
    ))
    try:
        r = requests.get(url)
    except Exception as err:
        raise Exception(err)
    if state in r.url:  # for safety check that 'state' is substring in response url
        bold = "\033[1m"
        end = "\033[0;0m"
        # copy url to the clipboard if we can
        # do not show user note if clipboard cmd is not set or executable is not found
        clip(r.url, show_note=False)
        print("1) Open following url in your browser.")
        print("2) If asked to login into twitch, you are required to do so, in order to get 'access_token' only known by twitch, and now also known by YOU! B)")
        print(f"{bold}After successful login, page is not existing! ALL WORK AS EXPECTED!{end}")
        print("3) Copy from browser url - part of 'access_token' content (from '=' to first '&' excluding those symbols!) and paste that as input here.")
        print(f"'{r.url}'")
        access_token = input("access_token=").strip()
        # try to get user_id by new access_token & validate that user put right access_token
        user_id = get_user_id(access_token, client_id)
        # write to private file for using in further requests
        write_private_data(user_id, access_token, client_id)
        print("SUCCESS")
    else:
        print(f"original state: '{state}' not matches state in response!")
        print("^ because of that - to prevent possible 'CSRF attack' on you, application was stopped!")
        exit(66)


if __name__ == "__main__":
    # execute when ran directly
    get_auth_token()



#!/usr/bin/env python3
# coding=utf-8

from twitchez import HEADER_H
from twitchez import STDSCR
from twitchez import __version__
from twitchez import keys
from twitchez import thumbnails

import curses
import re


def short_desc(string: str) -> str:
    """Make short readable description by removing one of specific patterns from string.
    Also replace _ characters by whitespace.
    """
    # Python 3.10.1 BUG:
    # >>> "tab_add".lstrip("tab_") => produces "dd" while it should be "add"
    patterns_to_strip = ["scroll_", "hint_", "tab", "bmark"]
    for pattern in patterns_to_strip:
        if string.startswith(pattern):
            string = string.lstrip(pattern)
            break  # remove only the first pattern found
    out = string.replace("_", " ").strip()
    return out


def table_lines(keysdict, header) -> list:
    """Generate simple list of strings with key and short description.
    Each line in the list is the same length (trailing whitespaces).
    """
    kws = 4        # num of ws after key char
    hws = kws + 1  # num of ws before header
    maxlen = 15
    for t in keysdict.keys():
        maxlen = max(maxlen, len(short_desc(t)))  # max length of longest line
    # header leading & trailing whitespaces
    hlws = " " * hws
    htws = " " * (maxlen - len(header) - len(hlws) + hws)
    outheader = hlws + header + htws
    list_of_lines = []
    list_of_lines.append(outheader)
    frmtstr = "{:" + str(kws) + "} {:" + str(maxlen) + "}"
    for name, key in keysdict.items():
        desc = short_desc(name)
        string = frmtstr.format(key, desc)
        list_of_lines.append(string)
    return list_of_lines


def append_blank_lines(table: list, num_of_out_lines: int) -> list:
    """Append to the table empty lines of the same max width till num of total lines.
    Used to make all tables of equal line count.
    """
    maxlen = len(max(table))  # max length of longest line (max len of element)
    blank_line = " " * maxlen
    while len(table) < num_of_out_lines:
        table.append(blank_line)
    return table


def simple_tables(area_width) -> tuple[int, str]:
    """Simple string tables as grid that fit in area_width.
    returns: total line count, multiline string as table.
    """
    sk = table_lines(keys.scroll_keys, "[SCROLL]")
    tk = table_lines(keys.tab_keys, "[TABS]")
    hk = table_lines(keys.hint_keys, "[HINTS]")
    ok = table_lines(keys.other_keys, "[OTHER]")
    bk = table_lines(keys.bmark_keys, "[BMARK]")
    tables = [sk, tk, hk, ok, bk]
    maxln = len(max(tables, key=len))  # max num of lines in table (max num of elements in list)
    maxlen = len(max(max(t, key=len) for t in tables))  # max length of longest line
    # make all tables of equal line count
    sk = append_blank_lines(sk, maxln)
    tk = append_blank_lines(tk, maxln)
    hk = append_blank_lines(hk, maxln)
    ok = append_blank_lines(ok, maxln)
    bk = append_blank_lines(bk, maxln)
    tables = [sk, tk, hk, ok, bk]

    # even spacing and indent from left
    maxcolnum = area_width // maxlen
    if maxcolnum > 5:  # limit max number of table columns
        maxcolnum = 5
    free_cols = area_width - int(maxlen * maxcolnum)
    if free_cols < maxcolnum or maxcolnum < 2:
        rem_on_col = 0
    else:
        rem_on_col = free_cols // maxcolnum - 1
    if rem_on_col == 0 or maxcolnum < 2:
        indentstr = ""
    else:
        # NOTE: the difference with one non-half indent is especially noticeable
        #       that the center is shifted at large terminal widths (200-239 cols)
        indentstr = " " * (rem_on_col // 2)
    strtemplateraw = indentstr + "{}" + indentstr

    out = ""
    add_row = "\n\n"  # new lines for the new row of tables
    # TODO: figure out how to make following code less ugly... (currently it is more like hardcoded)
    for num in range(len(tables), 0, -1):
        strtemplate = strtemplateraw * num
        if num == 5:
            out += "\n".join(strtemplate.format(t1, t2, t3, t4, t5) for t1, t2, t3, t4, t5 in zip(sk, tk, hk, ok, bk))
        elif num == 4:
            out += "\n".join(strtemplate.format(t1, t2, t3, t4) for t1, t2, t3, t4 in zip(sk, tk, hk, ok))
            out += add_row
            out += "\n".join(strtemplateraw.format(t5) for t5 in bk)
        elif num == 3:
            # FIXME
            out += "\n".join(strtemplate.format(t1, t2, t3) for t1, t2, t3 in zip(sk, tk, hk))
            out += add_row
            out += "\n".join((strtemplateraw * 2).format(t4, t5) for t4, t5 in zip(ok, bk))
        elif num == 2:
            out += "\n".join(strtemplate.format(t1, t2) for t1, t2 in zip(sk, tk))
            out += add_row
            out += "\n".join(strtemplate.format(t3, t4) for t3, t4 in zip(hk, ok))
            out += add_row
            out += "\n".join(strtemplateraw.format(t5) for t5 in bk)
        elif num == 1:
            for _t in tables:
                out += "\n".join(strtemplateraw.format(t1) for t1 in _t)
                out += add_row
        else:
            out = "E" * area_width
        out = out.rstrip()  # trim empty lines from the end of the out string & trailing ws
        maxlinelen = len(max(out.splitlines(), key=len))  # max length of longest line
        if maxlinelen <= area_width:
            break
        else:
            out = ""  # clear
    # replace repeating empty lines by a single empty line
    out = re.sub(r'\n\s*\n', '\n\n', out, re.MULTILINE)
    tln = out.count("\n") + 1  # total lines count
    return tln, out


def push_text(win, text: str, pos=0):
    """Add text str into window respecting it's height.
    Also update text after changing scroll position.
    (simple text string scrolling).
    """
    win.clear()
    h, _ = win.getmaxyx()
    lines = text.splitlines()
    if h >= len(lines):
        # all lines of text can simply be placed inside a win
        win.addstr(text)
    else:
        # addstr only slice of text that can fit inside a win
        text_slice = "\n".join(lines[pos:h + pos])
        win.addstr(text_slice)
    win.refresh()


def help():
    """Draw help window with key mappings and their description."""
    H, W = STDSCR.getmaxyx()
    if H < 10 or W < 20:
        return

    thumbnails.draw_stop()

    y, x = HEADER_H - 1, 2
    h, w = H - y * 2, W - x * 2

    close_help_keys = [
        curses.KEY_RESIZE,  # close on terminal resize event
        curses.KEY_F1,
        keys.other_keys.get("keys_help"),
        keys.other_keys.get("quit"),
    ]
    scroll_help_keys = keys.scroll_keys.values()

    v_str = f"v{__version__}"
    title = f" twitchez {v_str} "
    t_h_c = w // 2 - len(title) // 2  # title horizontal center

    win = STDSCR.derwin(h, w, y, x)
    win.clear()
    win.border()
    win.addstr(0, t_h_c, title, curses.A_BOLD)
    win.refresh()

    pad_y, pad_x = 2, 5
    pad_h = h - pad_y
    pad_w = w - pad_x

    pad = win.subpad(pad_h, pad_w, pad_y, pad_x)
    pad.scrollok(True)

    tln, table = simple_tables(pad_w)
    push_text(pad, table)

    end = tln - pad_h
    pos = 0

    while True:
        c = STDSCR.get_wch()
        # enable scrolling only if content doesn't fit in height entirely
        if tln > pad_h:
            if c in scroll_help_keys:
                if c == keys.scroll_keys.get("scroll_down"):
                    pos += 1
                elif c == keys.scroll_keys.get("scroll_up"):
                    pos -= 1
                elif c == keys.scroll_keys.get("scroll_down_page"):
                    pos += 5
                elif c == keys.scroll_keys.get("scroll_up_page"):
                    pos -= 5
                elif c == keys.scroll_keys.get("scroll_top"):
                    pos = 0
                elif c == keys.scroll_keys.get("scroll_bot"):
                    pos = end
                # limit scroll
                if pos < 0:
                    pos = 0
                elif pos > end:
                    pos = end
                push_text(pad, table, pos)
                continue
        if c in close_help_keys:
            pad.clear()
            pad.refresh()
            break


if __name__ == "__main__":
    def print_w_info(width):
        """To be able to see where are: width limit, center."""
        ln, table = simple_tables(width)
        fstring = " w:[{0}] ln:({1}) "
        info = fstring.format(width, ln)
        half = width // 2  # approx (as this is terminal cells)
        halfs = "─" * (half - 2)  # 2 = "x" as center + extra "─" in beg
        beg = "┌─" + info + halfs[len(info):]
        end = halfs + "┐"
        bar = beg + "x" + end
        print(bar)
        print(table)

    print("=" * 100)
    print_w_info(70)
    print_w_info(80)
    print_w_info(100)
    print_w_info(130)



#!/usr/bin/env python3
# coding=utf-8

from twitchez import command
from twitchez import conf

import subprocess
import os

try:
    import webbrowser
except ImportError:
    has_webbrowser = False
else:
    has_webbrowser = True


chat_cmd = conf.setting("open_chat_cmd")
executable = command.first_cmd_word(chat_cmd)
without_funcs = command.without_funcs(executable)
cmd_check = command.cmd_check(executable)


def open_chat(channel_login):
    """Open twitch chat of the channel."""
    url = f"https://www.twitch.tv/popout/{channel_login}/chat?popout="
    if cmd_check:
        cmd = chat_cmd.split()
        cmd.append(url)
        sub = subprocess.Popen
        sub(cmd,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL)
    elif has_webbrowser and not without_funcs:
        """Open in default browser using webbrowser module.
        Following (currently) are the only way to suppress stdout & stderr produced by webbrowser.open().
        We suppress everything for the case if webbrowser.open() outputs something before opening.
        Read more here: 'https://stackoverflow.com/a/2323563'.
        """
        savout = os.dup(1)  # stdout
        saverr = os.dup(2)  # stderr
        os.close(1)
        os.close(2)
        os.open(os.devnull, os.O_RDWR)
        try:
            webbrowser.open_new(url)
        finally:
            os.dup2(savout, 1)
            os.dup2(saverr, 2)


if __name__ == "__main__":
    open_chat("LIRIK")



#!/usr/bin/env python3
# coding=utf-8

from configparser import ConfigParser, NoSectionError
from pathlib import Path
from twitchez import TWITCHEZDIR, fs


def read_conf(*configs):
    """Read config files with omitted section.
    Last config takes precedence over previous configs.
    """
    parser = ConfigParser()
    section = "[GENERAL]\n"
    for f in configs:
        if Path(f).is_file():  # check that file exist
            with open(f) as stream:
                parser.read_string(section + stream.read())
    return parser


glob_conf_dir = Path(TWITCHEZDIR, "config").resolve()

glob_conf = Path(glob_conf_dir, "default.conf").resolve()
user_conf = Path(fs.get_user_conf_dir(), "config.conf").resolve()

glob_keys = Path(glob_conf_dir, "defkeys.conf").resolve()
user_keys = Path(fs.get_user_conf_dir(), "keys.conf").resolve()

temp_vars = Path(fs.get_tmp_dir(), "vars").resolve()
data_vars = Path(fs.get_data_dir("data"), "data").resolve()

config = read_conf(glob_conf, user_conf)
keymap = read_conf(glob_keys, user_keys)


def setting(keyname):
    found = config.get("GENERAL", keyname)
    return found


def key(keyname):
    found = keymap.get("GENERAL", keyname, fallback="")
    return found


def tmp_set(option, value, section="GENERAL"):
    """Set tmp variable value."""
    temp = ConfigParser()
    temp.read(temp_vars)
    if not temp.has_section(section):
        temp.add_section(section)
    temp.set(str(section), str(option), str(value))
    with open(temp_vars, "w") as f:
        temp.write(f, space_around_delimiters=False)


def tmp_get(keyname, fallback, section="GENERAL"):
    """Get tmp variable value."""
    temp = ConfigParser()
    temp.read(temp_vars)
    if not temp.has_section(section):
        temp.add_section(section)
    if fallback or not temp.has_option(section, keyname):
        found = temp.get(section, keyname, fallback=fallback)
    else:
        found = temp.get(section, keyname)
    return found


def cfpath(fallback: Path, fpath="") -> Path:
    """Return fallback path, if fpath is not valid."""
    ppath = Path(fpath)
    if not fpath or not ppath.is_file:
        ppath = fallback
    return ppath.resolve()


def dta_file(fpath="") -> Path:
    """Return fpath path or fallback to the default data file path."""
    return cfpath(data_vars, fpath)


def cp_dta(fpath: Path) -> tuple[ConfigParser, Path]:
    """Set defaults, read config & return class object."""
    fpath = dta_file(fpath.as_posix())
    dta = ConfigParser()
    # fix: preserve capitalization (option as is without transformation)
    # read more: https://docs.python.org/3/library/configparser.html#ConfigParser.optionxform(option)
    dta.optionxform = lambda option: option
    dta.read(fpath)
    return dta, fpath


def dta_set(option, value, section="GENERAL", fpath=""):
    """Set data variable value."""
    fpath = dta_file(fpath)
    dta, fp = cp_dta(fpath)
    if not dta.has_section(section):
        dta.add_section(section)
    dta.set(str(section), str(option), str(value))
    with open(fp, "w") as f:
        dta.write(f, space_around_delimiters=False)


def dta_get(option, fallback, section="GENERAL", fpath=""):
    """Get data variable value."""
    fpath = dta_file(fpath)
    dta, _ = cp_dta(fpath)
    if not dta.has_section(section):
        dta.add_section(section)
    if fallback or not dta.has_option(section, option):
        found = dta.get(section, option, fallback=fallback)
    else:
        found = dta.get(section, option)
    return found


def dta_rmo(option: str, section="GENERAL", fpath=""):
    """Remove data option."""
    fpath = dta_file(fpath)
    dta, fp = cp_dta(fpath)
    dta.remove_option(section, option)
    with open(fp, "w") as f:
        dta.write(f, space_around_delimiters=False)


def dta_list(section="GENERAL", fpath="") -> list:
    """Return a list of (name, value) tuples for each option in a section."""
    fpath = dta_file(fpath)
    dta, _ = cp_dta(fpath)
    try:
        return dta.items(section)
    except NoSectionError:
        return []



#!/usr/bin/env python3
# coding=utf-8


def check_auth_data():
    """Check user auth token before launch of the main program."""
    from twitchez import fs
    private_file = fs.private_data_path()
    # private_file is empty -> get auth token & write to private_file
    if not private_file.stat().st_size:
        from twitchez import auth
        auth.get_auth_token()
        print("Launch me again!")
        exit(69)


def main():
    check_auth_data()
    from twitchez import init
    init.main()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# coding=utf-8

import twitchez

from setuptools import find_packages, setup

import re


def remove_gif(text: str) -> str:
    return re.sub(r"\n^<img src=.*>$\n\n", "", text, flags=re.DOTALL | re.MULTILINE)


def remove_video(text: str) -> str:
    return re.sub(r"http.*\.mp4\n\n", "", text)


def replace_gh_video_via_mdlink(text: str) -> str:
    return re.sub(r"(http.*\.mp4)\n\n", r"### [CLICK TO WATCH DEMO VIDEO](\1)\n\n", text)


def clean_md(text: str) -> str:
    """Clean the markdown text for pypi.org.
    directly embedded github videos are not supported by pypi.org.
    """
    #  text = remove_gif(text)
    text = replace_gh_video_via_mdlink(text)
    return text


with open("README.md", "r") as f:
    long_description = f.read()
with open("requirements.txt", "r") as f:
    requirements = [line.strip() for line in f]

long_description = clean_md(long_description)

setup(
    name="twitchez",
    version=twitchez.__version__,
    license=twitchez.__license__,
    url=twitchez.__url_project__,
    author=twitchez.__author__,
    description=twitchez.__description__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    python_requires=">=3.6",
    install_requires=requirements,
    extras_require={
        "thumbnails": ["ueberzugpp", "ueberzug"],
    },
    package_data={
        "twitchez": ["config/*.conf", "config/blank.jpg"],
    },
    entry_points={
        "console_scripts": ["twitchez=twitchez.__main__:main"]
    },
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
        "Environment :: Console :: Curses",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Development Status :: 4 - Beta",
    ],
    keywords="twitch TUI terminal curses ui client thumbnail image twitch.tv",
    project_urls={
        "Bug Reports": twitchez.__url_bug_reports__,
        "Source": twitchez.__url_repository__,
    },
)



#!/usr/bin/env python3
# coding=utf-8

from shutil import which
from twitchez import ENCODING
from twitchez import command
from twitchez import conf
import subprocess

notify_cmd = conf.setting("notify_cmd")
executable = command.first_cmd_word(notify_cmd)
without_funcs = command.without_funcs(executable)
cmd_check = command.cmd_check(executable)


def expire_time():
    """return expire time for notifications from config."""
    return conf.setting("notify_time")


def dunstify_cmd() -> list:
    # NOTE: dunst stack tag 'hi' stands for 'history ignore' and can be used for that purpose.
    # (requires creating matching rule in dunst config)
    t = expire_time()
    DST = "string:x-dunst-stack-tag"
    cmd = f"dunstify -t {t} -u low -h {DST}:twitchez -h {DST}:hi"
    return cmd.split()


def notify_send_cmd() -> list:
    t = expire_time()
    DST = "string:x-dunst-stack-tag"
    cmd = f"notify-send -t {t} -u low -h {DST}:twitchez -h {DST}:hi"
    return cmd.split()


def raise_user_note():
    """raise exception for regular user without traceback."""
    if without_funcs:
        return
    a = "A program to send desktop notifications was not found at your 'PATH'."
    b = "You can install 'notify-send' and it will be working by default."
    c = "Also you can set your own program cmd via 'notify_cmd = your cmd' in config."
    d = "If you want to use this program without seeing any notifications from it,"
    e = "simply paste next line in your config:"
    f = "notify_cmd = false"
    full_text = f"\n{a}\n{b}\n{c}\n{d}\n{e}\n{f}\n"
    raise Exception(full_text)


def get_notify_cmd(show_note: bool) -> list:
    """Check & return cmd if executable is on PATH."""
    cmd = []
    # prefer notify_cmd if set in config and executable found at PATH
    if cmd_check:
        cmd = notify_cmd.split()
    elif which("dunstify"):
        cmd = dunstify_cmd()
    elif which("notify-send"):
        cmd = notify_send_cmd()
    else:
        if show_note:
            raise_user_note()
        else:
            return []
    return cmd


def notify(body="", summary="", error=False, show_note=True):
    """Show user notification."""
    if without_funcs:
        return
    cmd = get_notify_cmd(show_note=show_note)
    if not cmd:
        return
    s = f"{summary}"
    b = f"{body}"
    # NOTE: if user specified custom notify_cmd that does not support additional args
    # => cmd will break. Because we append variable amount of options after getting cmd.
    if error:
        cmd.append("-u")
        cmd.append("critical")
        cmd.append("-t")
        cmd.append("8000")
    cmd.append(s)
    cmd.append(b)
    p = subprocess.Popen(cmd, text=True, encoding=ENCODING)
    p.wait()  # wait for process to finish



#!/usr/bin/env python3
# coding=utf-8

from ast import literal_eval
from pathlib import Path
from twitchez import conf
from twitchez import fs
from twitchez import paged
from twitchez.iselect import iselect

FILE = Path(fs.get_data_dir("data"), "tabs").resolve().as_posix()
LTABS = "LTABS"
DTABS = "DTABS"


def tabs_list() -> list:
    """Return list of tabs (name, dict) tuples."""
    return conf.dta_list(DTABS, FILE)


def tabs_dnames() -> list:
    """Return list of tab names (keys) of the dictionaries."""
    return [tname for tname, _ in tabs_list()]


def tab_names_ordered() -> list:
    """Return an ordered list of opened tab names."""
    names_list_str = conf.dta_get("ltabs", "", LTABS, FILE)
    try:
        names_list = literal_eval(names_list_str)
    except Exception as e:
        raise ValueError(f"names_list_str: '{names_list_str}'\n{e}")
    return names_list


def tabs_upd(tabs: list):
    """Update/set list of tabs."""
    conf.dta_set("ltabs", tabs, LTABS, FILE)
    # remove all not opened tabs (real tabs data with dicts)
    for tname in tabs_dnames():
        if tname not in tabs:
            conf.dta_rmo(tname, DTABS, FILE)


def cpname_set(pname: str):
    """Set value of the current page/tab name."""
    conf.dta_set("cpname", pname, LTABS, FILE)


def tab_upd(page_name: str, page_dict: dict):
    """Update tabs, set current page name, add page to the tabs list (if not exist)."""
    cpname_set(page_name)  # set the new current page name (NOTE: before everything else!)
    conf.dta_set(page_name, page_dict, DTABS, FILE)
    tab_add_new(page_name)


def cpname() -> str:
    """Get current page name, set to the first tab name as fallback."""
    cpn = conf.dta_get("cpname", "", LTABS, FILE)
    tabs = tab_names_ordered()
    if not cpn or cpn not in tabs:
        if not tabs:
            cpn = paged.FLPN  # fallback to following live page
        else:
            cpn = tabs[0]  # set first tab name as the current page name
        cpname_set(cpn)
    return cpn


def cpdict() -> dict:
    """Get current page dict."""
    return pdict(cpname())


def pdict(page_name="") -> dict:
    """Return page dict by the page name or (current tab/page by default)."""
    if not page_name:  # return page_dict of current tab/page
        pdict_str = conf.dta_get(cpname(), "", DTABS, FILE)
    else:
        pdict_str = conf.dta_get(page_name, cpname(), DTABS, FILE)
    if not pdict_str or pdict_str == paged.FLPN or page_name == paged.FLPN:
        return paged.following_live()  # fallback to following live page
    try:
        page_dict = literal_eval(pdict_str)
    except Exception as e:
        raise ValueError(f"pdict_str: '{pdict_str}'\n{e}")
    return page_dict


def tab_add_new(page_name: str):
    """Add new tab/page to the tabs list (if not exist)."""
    tabs = tab_names_ordered()
    if not tabs:
        tabs.append(page_name)
        tabs_upd(tabs)
        cpn = cpname()
        return
    # do not add the same tab twice
    if page_name not in tabs:
        cpn = cpname()
        # find index of current page name and insert new tab after that index
        cindex = tabs.index(cpn)
        nindex = cindex + 1
        tabs.insert(nindex, page_name)
        tabs_upd(tabs)
        return


def delete_tab(page_name="") -> dict:
    """Delete tab by page name or current tab/page and return page_dict of the previous tab."""
    ctab = cpname()
    tabs = tab_names_ordered()
    if (page_name != ctab and page_name in tabs):
        tab_to_delete = page_name
        tab_to_jump = ctab
    else:
        tab_to_delete = ctab
        _, tab_to_jump = prev_tab()

    if (tab_to_delete in tabs):
        tabs.remove(tab_to_delete)

    tabs_upd(tabs)
    return pdict(tab_to_jump)


def find_tab(fallback=cpdict()) -> dict:
    """Find and return page dict of selected tab or fallback to current tab (by default)."""
    tabs = tab_names_ordered()
    mulstr = "\n".join(tabs)  # each list element on it's own line
    tabname = iselect(mulstr, 130)
    # handle cancel of the command
    if tabname == 130:
        return fallback
    return pdict(tabname)


def next_tab() -> tuple[dict, str]:
    """Return (page_dict, page_name) tuple of the next tab (carousel)."""
    tabs = tab_names_ordered()
    cindex = tabs.index(cpname())
    nindex = cindex + 1
    if nindex > len(tabs) - 1:
        ntabname = tabs[0]
    else:
        ntabname = tabs[nindex]
    return pdict(ntabname), ntabname


def prev_tab() -> tuple[dict, str]:
    """Return (page_dict, page_name) tuple of the prev tab (carousel)."""
    tabs = tab_names_ordered()
    cindex = tabs.index(cpname())
    pindex = cindex - 1
    ptabname = tabs[pindex]
    return pdict(ptabname), ptabname



#!/usr/bin/env python3
# coding=utf-8

from twitchez import command
from twitchez import conf
from twitchez import fs
from twitchez import utils

from pathlib import Path
from shutil import which
from sys import version_info, stdout
from threading import Timer

import aiohttp
import asyncio
import json
import os                       # listdir, sep, devnull, basename, splitext
import subprocess


# check if executables at PATH
HAS_UEBERZUG = bool(which("ueberzugpp")) | bool(which("ueberzug"))
# also check user cmd in case executable provided via full path
HAS_UEBERZUG |= command.conf_cmd_check("ueberzug_cmd")[0]


def raise_user_note():
    """raise exception for regular user without traceback."""
    raise Exception(
        "\n\n"
        "Neither ueberzugpp nor ueberzug were found at PATH. While text_mode is not enabled.\n"
        "You can install 'ueberzugpp' and it will be working by default.\n"
        "It supports more output options and platforms than old version of ueberzug written in python.\n"
        "Also you can set your own program cmd via 'ueberzug_cmd = your cmd' in config.\n"
        "This will allow you to override default ueberzugpp layer --output for you machine, etc.\n"
        "If you want to use this program without thumbnails,\n"
        "simply paste next line in your config:\n"
        "text_mode = 3\n"
    )


def get_ueberzug_cmd() -> list:
    """Check & return cmd if executable is on PATH."""
    cmd = ""
    user_cmd_ok, ueberzug_cmd = command.conf_cmd_check("ueberzug_cmd")
    if user_cmd_ok:
        # prefer ueberzug_cmd if set in config and found at PATH or if full path provided
        cmd = ueberzug_cmd
    elif which("ueberzugpp"):
        # NOTE: --no-cache is important, without it ueberzugpp will show old cached thumbnails!
        cmd = "ueberzugpp layer --no-cache --silent"
    elif which("ueberzug"):
        # seems like python version of ueberzug have no layer specific options
        cmd = "ueberzug layer"
    else:
        raise_user_note()
    return cmd.split()


def text_mode() -> int:
    """Text mode: 0 => thumbnails mode (min: 0, max: 3).
    [1-3] => do not do anything with thumbnails do not even download them!
    The higher the value, the more rows of cells there will be in the grid.
    """
    tm = int(conf.setting("text_mode"))
    if tm < 0:
        tm = 0
    elif tm > 3:
        tm = 3
    # explicit text mode if ueberzug not found (optional dependency)
    if not HAS_UEBERZUG and tm < 1:
        tm = 1
    return tm


def rdiv() -> int:
    """Thumbnail resolution divisor (min: 2, max: 10)."""
    div = 1 + int(conf.setting("grid_size"))
    if div < 2:
        div = 2
    elif div > 10:
        div = 10
    return div


def container_size(thumbnail=False) -> tuple[int, int]:
    """Return tuple: (width, height) - based on divisor key in table.
    Selected values are close as possible to the real resolution of the thumbnails.
    Except couple values where visual result more appropriate: with (2,3,4) as divisor.
    """
    table = {
        10: (24, 7),
        9: (27, 8),
        8: (30, 9),
        7: (35, 10),
        6: (40, 11),
        5: (48, 13),
        4: (56, 15),
        3: (76, 20),
        2: (90, 24),
    }
    # use fallback key if div key not found
    _def_fix: tuple = (40, 11)  # fix: None is not assignable
    w, h = tuple(table.get(rdiv(), table.get(6, _def_fix)))
    # width/height modifier for perfect placement of thumbnails in the grid (very font dependent)
    w += int(conf.setting("wmod"))
    h += int(conf.setting("hmod"))
    tm = text_mode()
    if tm:
        return w, h - tm
    elif thumbnail:
        return w, h
    else:
        NLC = 3  # num of content lines in the box
        return w, h + NLC


def thumbnail_resolution() -> tuple[int, int]:
    """Return tuple: (width, height) - based on divisor key in table.
    really simple:  divisor = 10
    (1920, 1080) / 10 = (192, 108)
    The only values that don't match the actual result: divisor=2.
    """
    table = {
        10: (192, 108),
        9: (213, 120),
        8: (240, 135),
        7: (274, 154),
        6: (320, 180),
        5: (384, 216),
        4: (480, 270),
        3: (640, 360),
        2: (720, 405),  # actual (960, 540) is overkill!
    }
    # use fallback key if div key not found
    _def_fix: tuple = (320, 180)  # fix: None is not assignable
    return table.get(rdiv(), table.get(6, _def_fix))


def get_thumbnail_urls(rawurls) -> list:
    """Return thumbnail urls with {width} and {height} replaced."""
    width, height = thumbnail_resolution()
    urls = []
    for url in rawurls:
        # fix: video thumbnails currently have weird format with % characters
        if url and "%{" in url:
            # remove % character from thumbnail url
            url = url.replace("%{", "{")
        if not url:
            url = ""
        else:
            url = url.format(width=width, height=height)
        urls.append(url)
    return urls


async def fetch_image(session, url):
    """Asynchronously fetch image from url."""
    if not url:
        return None
    async with session.get(url) as response:
        return await response.read()


async def get_thumbnails_async(ids: list, rawurls: list, *subdirs) -> dict:
    """Asynchronously download thumbnails and return paths.
    (Actual realization)
    """
    thumbnail_paths = {}
    urls = get_thumbnail_urls(rawurls)
    tmpd = fs.get_tmp_dir("thumbnails", *subdirs)
    blank_thumbnail = Path(conf.glob_conf_dir, "blank.jpg")
    tasks = []
    async with aiohttp.ClientSession() as session:
        for url in urls:
            tasks.append(fetch_image(session, url))
        # wait until all thumbnails with non empty url are fetched
        thumbnails = await asyncio.gather(*tasks)

    for tid, thumbnail in zip(ids, thumbnails):
        thumbnail_fname = f"{tid}.jpg"
        thumbnail_path = Path(tmpd, thumbnail_fname)
        if thumbnail is None:
            if thumbnail_path.is_file() and thumbnail_path.samefile(blank_thumbnail):
                pass
            else:
                # remove symlink or file before creating new symlink
                if thumbnail_path.is_file():
                    thumbnail_path.unlink(missing_ok=True)
                # create symlink of blank_thumbnail
                thumbnail_path.symlink_to(blank_thumbnail)
        else:
            # NOTE: if existing thumbnail_path is symlink, original blank thumbnail
            # will be replaced by the thumbnail_path image, to prevent that
            # => remove symlink before writing new image file
            if thumbnail_path.is_symlink():
                thumbnail_path.unlink(missing_ok=True)
            with open(thumbnail_path, 'wb') as f:
                f.write(thumbnail)
        thumbnail_paths[tid] = str(thumbnail_path)
    return thumbnail_paths


def download_thumbnails(ids: list, rawurls: list, *subdirs) -> dict:
    """Asynchronously download thumbnails and return paths.
    (Wrapper with asyncio run/run_until_complete)
    """
    if version_info >= (3, 7):  # Python 3.7+
        return asyncio.run(get_thumbnails_async(ids, rawurls, *subdirs))
    else:  # Python 3.5-3.6
        loop = asyncio.get_event_loop()
        try:
            return loop.run_until_complete(get_thumbnails_async(ids, rawurls, *subdirs))
        finally:
            loop.close()


def find_thumbnails(ids: list, *subdirs) -> dict:
    """Find and return previously downloaded thumbnails paths."""
    tmpd = fs.get_tmp_dir("thumbnails", *subdirs)
    blank_thumbnail = Path(conf.glob_conf_dir, "blank.jpg")

    tnames = utils.replace_pattern_in_all(os.listdir(tmpd), ".jpg", "")
    differ = list(set(tnames).difference(set(ids)))
    fnames = utils.add_str_to_list(differ, ".jpg")  # add file extension back
    for fname in fnames:
        # remove thumbnail files/symlinks which id not in ids list
        Path(tmpd, fname).unlink(missing_ok=True)

    thumbnail_list = utils.insert_to_all(os.listdir(tmpd), tmpd, opt_sep=os.sep)
    thumbnail_paths = {}
    for path in thumbnail_list:
        tid = os.path.basename(os.path.splitext(path)[0])  # file basename without .ext
        thumbnail_paths[tid] = path
    # fix: if thumbnail_paths does not have id from ids
    # this usually happens if text mode without thumbnails was previously set
    for tid in ids:
        if tid not in thumbnail_paths.keys():
            thumbnail_paths[tid] = str(blank_thumbnail)
    return thumbnail_paths


class Thumbnail:
    """Prepare Thumbnail ueberzug parameters and add to Thumbnails."""
    w, h = container_size(thumbnail=True)

    def __init__(self, identifier, img_path, x, y):
        self.identifier = identifier
        self.img_path = img_path
        self.x = x
        self.y = y
        self.ue_params = self._ue_params()

    def _ue_params(self) -> dict[str, str]:
        """Return dict for thumbnail with all parameters required by ueberzug.
        Append parameters of the Thumbnail to the list of Thumbnails parameters.
        """
        uep = {
            "action": "add",
            "scaler": "fit_contain",
            "identifier": self.identifier,
            "path": self.img_path,
            "height": self.h,
            "width": self.w,
            "y": self.y,
            "x": self.x,
        }
        if not text_mode():
            Thumbnails.uepl.append(uep)
        return uep


class Thumbnails:
    uepl: list[dict[str, str]] = []  # ueberzug list of thumbnail parameters
    tm = text_mode()

    is_initialized = False
    working_dir = fs.get_tmp_dir()

    @staticmethod
    def json_schema_thumbnails(uepl: list) -> str:
        """New Line Delimited JSON, one ueberzug thumbnail parameters data per line."""
        nl_json = ""
        for th_params in uepl:
            nl_json += json.dumps(th_params) + '\n'
        return nl_json

    @staticmethod
    def PopenType() -> subprocess.Popen:
        """Get proper type and object properties at initialization.
        By executing common/smallest/fastest program existing in nearly every OS.
        """
        return subprocess.Popen(["echo"], stdin=subprocess.PIPE, stdout=subprocess.DEVNULL)

    def __init__(self):
        self.sub_proc = self.PopenType()

    def init_check(self) -> bool:
        return self.is_initialized and self.sub_proc.poll() is None

    def init(self):
        """start ueberzug subprocess."""
        assert (self.sub_proc.stdin is not None)  # "None" [reportOptionalMemberAccess]
        if (self.init_check()):
            return
        cmd: list = get_ueberzug_cmd()
        # we do not want to close subprocess because that stops the drawing.
        with open(os.devnull, "wb", 0) as devnull:
            self.sub_proc = subprocess.Popen(
                cmd,
                cwd=self.working_dir,
                stderr=devnull,
                stdout=stdout.buffer,
                stdin=subprocess.PIPE,
                universal_newlines=True,
            )
        self.is_initialized = True

    def execute(self, **kwargs):
        """execute ueberzug action/cmd."""
        self.init()
        assert (self.sub_proc.stdin is not None)  # "None" [reportOptionalMemberAccess]
        # NOTE: direct interaction with the stdin as we do not want to close subprocess.
        if kwargs:
            # NOTE: mainly for the cleanup (remove action)
            self.sub_proc.stdin.write(json.dumps(kwargs) + '\n')
        else:
            self.sub_proc.stdin.write(self.json_schema_thumbnails(self.uepl))
        self.sub_proc.stdin.flush()

    def clear(self):
        """cleanup from the old thumbnails data."""
        assert (self.sub_proc.stdin is not None)  # "None" [reportOptionalMemberAccess]
        if self.sub_proc and not self.sub_proc.stdin.closed:
            for th_params in self.uepl:
                # remove previously added but no longer needed thumbnails
                self.execute(action="remove", identifier=th_params["identifier"])
        # clear list of the thumbnails parameters
        self.uepl.clear()

    def quit(self):
        """wrapper for the fast & safe termination of subprocess."""
        if self.init_check():
            timer_kill = Timer(1, self.sub_proc.kill, [])
            try:
                self.sub_proc.terminate()
                timer_kill.start()
                self.sub_proc.communicate()
            finally:
                timer_kill.cancel()

    def start(self):
        """Start drawing images via subprocess."""
        if self.tm:
            return
        self.execute()

    def finish(self, safe=False):
        """Finish drawing images and optionally terminate subprocess."""
        if self.tm:
            return
        self.clear()
        if safe:
            self.quit()


THUMBNAILS: Thumbnails = Thumbnails()


def draw_start():
    if not HAS_UEBERZUG:
        return
    THUMBNAILS.start()


def draw_stop(safe=False):
    if not HAS_UEBERZUG:
        return
    THUMBNAILS.finish(safe)



#!/usr/bin/env python3
# coding=utf-8

from twitchez import STDSCR
from datetime import datetime
from difflib import SequenceMatcher
from os.path import getmtime
from re import compile
from twitchez import conf
import textwrap
import time


# visible length of one emoji in terminal cells
EMOJI_CELLS = int(conf.setting("emoji_cells"))

EMOJI_PATTERN = compile(
    "["
    "\U0001F1E0-\U0001F1FF"  # flags (iOS)
    "\U0001F300-\U0001F5FF"  # symbols & pictographs
    "\U0001F600-\U0001F64F"  # emoticons
    "\U0001F680-\U0001F6FF"  # transport & map symbols
    "\U0001F700-\U0001F77F"  # alchemical symbols
    "\U0001F780-\U0001F7FF"  # Geometric Shapes Extended
    "\U0001F800-\U0001F8FF"  # Supplemental Arrows-C
    "\U0001F900-\U0001F9FF"  # Supplemental Symbols and Pictographs
    "\U0001FA00-\U0001FA6F"  # Chess Symbols
    "\U0001FA70-\U0001FAFF"  # Symbols and Pictographs Extended-A
    "\U00002702-\U000027B0"  # Dingbats
    "\U000024C2-\U0001F251"
    "]+"
)


def tryencoding(string: str) -> str:
    """Return string in default encoding or
    if not printable -> try to re-encode into utf-16."""
    if not string.isprintable():
        try:
            string = string.encode('utf-16', 'surrogatepass').decode("utf-16", "ignore")
        except (UnicodeEncodeError, UnicodeDecodeError) as e:
            string = str(e)
    return string


def demojize(str: str) -> str:
    """Return string without emojis."""
    return EMOJI_PATTERN.sub(r'', str)


def emoji_count(str: str) -> int:
    """Returns the count of emojis in a string."""
    return len(str) - len(demojize(str))


def tlen(str: str) -> int:
    """Return len of str respecting emoji visible length in terminal cells.
    EMOJI_CELLS: visible length of one emoji in terminal cells.
    """
    if EMOJI_CELLS < 2:
        return len(str)
    else:
        return EMOJI_CELLS * emoji_count(str) + len(demojize(str))


def was_resized(xysum=0) -> int:
    """HACK: to be able to check if terminal was resized
    between calls of the function inside some function (long running).
    Compares sum of xy between the initial
    and the actual xy sum at the call time of the function.
    """
    def sumyx() -> int:
        return sum(STDSCR.getmaxyx())  # sum of tuple[x, y]

    def was_resized_between_calls() -> int:
        return 0 if xysum == sumyx() else 1

    # just return the initial xysum at the call time
    if xysum == 0:
        return sumyx()
    else:
        return was_resized_between_calls()


def secs_since_mtime(path):
    """time_now - target_mtime = int(secs)."""
    return int(time.time() - getmtime(path))


def replace_pattern_in_all(inputlist, oldstr, newstr) -> list:
    """Replace oldstr with newstr in all items from a list."""
    outputlist = []
    for e in inputlist:
        outputlist.append(str(e).replace(oldstr, newstr))
    return outputlist


def add_str_to_list(input_list, string) -> list:
    """Add string to the end of all elements in a list."""
    outputlist = [e + str(string) for e in input_list]
    return outputlist


def insert_to_all(list, string, opt_sep="") -> list:
    """ Insert the string at the beginning of all items in a list. """
    string = str(string)
    if opt_sep:
        string = f"{string}{opt_sep}"
    string += '% s'
    list = [string % i for i in list]
    return list


def strws(str: str) -> str:
    """Return a str without whitespaces & slash characters - replaced by '_'."""
    return str.strip().replace(' ', '_').replace('/', '_').replace('\\', '_')


def strclean(str: str) -> str:
    """return slightly cleaner string."""
    # remove unneeded characters from string
    s = str.replace("\n", " ").replace("\t", " ")
    # replace repeating whitespaces by single whitespace
    s = ' '.join(s.split())
    s = s.strip()
    return s


def strtoolong(str: str, width: int, indicator="..") -> str:
    """Return str slice of width with indicator at the end.
    (to show that the string cannot fit completely in width)
    """
    if tlen(str) > width:
        str_fit_in_width = str[:width]
        # visible width in terminal cells that str occupies
        terminal_cells = tlen(str_fit_in_width)
        if terminal_cells > width:
            ec = emoji_count(str_fit_in_width)
            cut = ec + len(indicator)
            out_str = str_fit_in_width[:-cut] + indicator
        else:
            out_str = str_fit_in_width[:-len(indicator)] + indicator
        return out_str
    else:
        return str


def word_wrap_title(string: str, width: int, max_len: int, max_lines=3) -> str:
    """Word wrap title string."""
    string = strclean(string)
    if tlen(string) <= width:
        return string
    title_lines = textwrap.wrap(
        string, width, max_lines=max_lines,
        expand_tabs=False, replace_whitespace=True,
        break_long_words=True, break_on_hyphens=True, drop_whitespace=True
    )
    out_str = ""
    cline = 0
    for line in title_lines:
        cline += 1
        if len(line) == width:
            out_str += line
        else:
            out_str += f"{line}\n"
    # limit string len
    if len(out_str) > max_len:
        out_str = out_str[:max_len]
    # add mask only if length of last line met condition
    if len(title_lines[-1]) < width // 2:
        mask = "  "  # mask to differentiate from underlying text
        out_str = out_str[:-len(mask) + 1] + mask
    return out_str


def sdate(isodate: str) -> str:
    """Take iso date str and return shorten date str."""
    # remove Z character from default twitch date (2021-12-08T11:43:43Z)
    idate = isodate.replace("Z", "")
    vdate = datetime.fromisoformat(idate).isoformat(' ', 'minutes')
    today = datetime.today().isoformat(' ', 'minutes')
    current_year = today[:4]
    if current_year not in vdate:
        pattern = vdate[-6:]  # cut off only time
    else:
        sm = SequenceMatcher(None, vdate, today)
        match = sm.find_longest_match(0, len(vdate), 0, len(today))
        # longest common string between two
        pattern = vdate[match.a: match.a + match.size - 1]
    # remove pattern, cut leading '-' and strip whitespaces
    sdate = str(vdate).replace(pattern, "").strip("-").strip()
    return sdate


def duration(duration: str, simple=False, noprocessing=False) -> str:
    """Take twitch duration str and return duration with : as separators.
    Can optionally return a str without processing or with simple str processing.
    """
    if noprocessing:
        return duration
    if simple:
        # downside is very variable length of str and subjective ugliness of result.
        return duration.replace("h", ":").replace("m", ":").replace("s", ":").strip(":")
    # Don't see any real benefit of the following code over a silly simple one-liner :)
    # Result of the following algorithm are prettier, but also produces longer str.
    if "h" in duration:
        # extract hours from string
        H, _, _ = duration.partition("h")
        H = int(H.strip())
        # fix: if hours > 23 => put hours as simple str into format
        if H > 23:
            ifmt = f"{H}h%Mm%Ss"
            ofmt = f"{H}:%M:%S"
        else:
            ifmt = "%Hh%Mm%Ss"
            ofmt = "%H:%M:%S"
    elif "m" in duration:
        ifmt = "%Mm%Ss"
        ofmt = "%M:%S"
    else:
        return duration
    idur = datetime.strptime(duration, ifmt)
    odur = str(idur.strftime(ofmt))
    return odur
